/* SCR2PC - ZX Spectrum 2 ColecoVision BITMAP Converter
 * ----------------------------------------------------
 * HISTORY
 * 20170108 - ALPHA - FIRST DRAFT
 */
#include <stdio.h> /* printf, file */
#include <stdlib.h> /* exit */
#include <unistd.h> /* getopt */
#include <string.h> /* memcpy, memset */

/*
 * - PROGRAM BASIC AUTHOR AND COPYRIGHTS INFORMATIONS -
 */
#define TITLE "SCR2PC ZX Spectrum to ColecoVision BITMAP"
#define VERSION 0
#define SUBVERSION 1
#define DESCRIPTION "Convert ZX Spectrum (.scr .mlt) to ColecoVision (.pc)"
#define AUTHOR "Daniel Bienvenu aka NewColeco"
#define YEAR 2017

/*
 * - BOOLEAN VALUES -
 */
#define TRUE -1
#define FALSE 0

/*
 * - PROGRAM VARIABLES -
 */
int bYes;


/*
 * - FILE STRUCTS -
 */
struct ZX_SCR {
	unsigned char PATTERN[6144];
	unsigned char COLOR[768];
} zx_bitmap_scr;

struct ZX_MLT {
	unsigned char PATTERN[6144];
	unsigned char COLOR[6144];
} zx_bitmap_mlt;

struct COLECO_PC {
	unsigned char PATTERN[6144];
	unsigned char COLOR[6144];
} cv_bitmap;

/*
 * - COLOR PALETTE CONVERSION : ZX TO CV -
 */
static const unsigned char ZX_TO_COLECO_COLOR[] =
{
	/* BRIGHT 0 */
	0x11, 0x41, 0x81, 0xd1, 0x21, 0x71, 0xa1, 0xe1,
	0x14, 0x44, 0x84, 0xd4, 0x24, 0x74, 0xa4, 0xe4,
	0x18, 0x48, 0x88, 0xd8, 0x28, 0x78, 0xa8, 0xe8,
	0x1d, 0x4d, 0x8d, 0xdd, 0x2d, 0x7d, 0xad, 0xed,
	0x12, 0x42, 0x82, 0xd2, 0x22, 0x72, 0xa2, 0xe2,
	0x17, 0x47, 0x87, 0xd7, 0x27, 0x77, 0xa7, 0xe7,
	0x1a, 0x4a, 0x8a, 0xda, 0x2a, 0x7a, 0xaa, 0xea,
	0x1e, 0x4e, 0x8e, 0xde, 0x2e, 0x7e, 0xae, 0xee,
	/* BRIGHT 1 */
	0x11, 0x51, 0x91, 0xd1, 0x31, 0x71, 0xb1, 0xf1,
	0x15, 0x55, 0x95, 0xd5, 0x35, 0x75, 0xb5, 0xf5,
	0x19, 0x59, 0x99, 0xd9, 0x39, 0x79, 0xb9, 0xf9,
	0x1d, 0x5d, 0x9d, 0xdd, 0x3d, 0x7d, 0xbd, 0xfd,
	0x13, 0x53, 0x93, 0xd3, 0x33, 0x73, 0xb3, 0xf3,
	0x17, 0x57, 0x97, 0xd7, 0x37, 0x77, 0xb7, 0xf7,
	0x1b, 0x5b, 0x9b, 0xdb, 0x3b, 0x7b, 0xbb, 0xfb,
	0x1f, 0x5f, 0x9f, 0xdf, 0x3f, 0x7f, 0xbf, 0xff,
	/* BRIGHT 0 */
	0x11, 0x41, 0x81, 0xd1, 0x21, 0x71, 0xa1, 0xe1,
	0x14, 0x44, 0x84, 0xd4, 0x24, 0x74, 0xa4, 0xe4,
	0x18, 0x48, 0x88, 0xd8, 0x28, 0x78, 0xa8, 0xe8,
	0x1d, 0x4d, 0x8d, 0xdd, 0x2d, 0x7d, 0xad, 0xed,
	0x12, 0x42, 0x82, 0xd2, 0x22, 0x72, 0xa2, 0xe2,
	0x17, 0x47, 0x87, 0xd7, 0x27, 0x77, 0xa7, 0xe7,
	0x1a, 0x4a, 0x8a, 0xda, 0x2a, 0x7a, 0xaa, 0xea,
	0x1e, 0x4e, 0x8e, 0xde, 0x2e, 0x7e, 0xae, 0xee,
	/* BRIGHT 1 */
	0x11, 0x51, 0x91, 0xd1, 0x31, 0x71, 0xb1, 0xf1,
	0x15, 0x55, 0x95, 0xd5, 0x35, 0x75, 0xb5, 0xf5,
	0x19, 0x59, 0x99, 0xd9, 0x39, 0x79, 0xb9, 0xf9,
	0x1d, 0x5d, 0x9d, 0xdd, 0x3d, 0x7d, 0xbd, 0xfd,
	0x13, 0x53, 0x93, 0xd3, 0x33, 0x73, 0xb3, 0xf3,
	0x17, 0x57, 0x97, 0xd7, 0x37, 0x77, 0xb7, 0xf7,
	0x1b, 0x5b, 0x9b, 0xdb, 0x3b, 0x7b, 0xbb, 0xfb,
	0x1f, 0x5f, 0x9f, 0xdf, 0x3f, 0x7f, 0xbf, 0xff
};

void scr2mlt()
{
	int i,j;
	unsigned char c;
	unsigned char *pointer_source = &zx_bitmap_scr.COLOR[0];
	unsigned char *pointer_destination = &zx_bitmap_mlt.COLOR[0];
	for (j=0;j<24;j++)
	{
		for (i=0;i<8;i++)
		{
			memcpy(pointer_destination, pointer_source, 32);
			pointer_destination += 32;
		}
		pointer_source += 32;
	}
	memcpy(zx_bitmap_mlt.PATTERN,zx_bitmap_scr.PATTERN,sizeof(zx_bitmap_mlt.PATTERN));
}

void mlt2cv()
{
	int i,j,k,m;
	int cv_offset_c,cv_offset_p,zx_offset;
	zx_offset = cv_offset_c = cv_offset_p = 0; 
	for (k=0;k<3;k++)
	{
		for (j=0;j<8;j++)
		{
			for (i=0;i<8;i++)
			{
				for (m=0;m<32;m++)
				{
					cv_bitmap.PATTERN[cv_offset_p] = zx_bitmap_mlt.PATTERN[zx_offset];
		   	   	   	cv_bitmap.COLOR[cv_offset_c] = ZX_TO_COLECO_COLOR[zx_bitmap_mlt.COLOR[zx_offset]];
					cv_offset_p +=8;
					cv_offset_c +=8;	   	   	   	   	   
					zx_offset++;
				}
				cv_offset_c -= (256 - 1);
			}
			cv_offset_p -= (256*8 - 1);
			cv_offset_c -= (8 - 256);
		}
		cv_offset_p += (256*8 - 8);
	}
}

void help(void)
{
   	printf("\nFor help type : pcspliter -h\n");
	printf("Press Enter to exit.\n");
	getchar();
	exit(0);
}

int file_exits(char *filename)
{
	int exists;
	FILE *file = NULL;
	file = fopen(filename,"r");
	exists = (file != NULL);
	fclose(file);
	return(exists);	   
}

long file_size(FILE *file)
{
	long size;
	fseek(file,0l,SEEK_END);
	size = ftell(file);
	fseek(file,0l,SEEK_SET);
	/* rewind(file);*/
	return size;
}

int yesno ()
{
	char cYesNo;
	do {
		scanf(" %c",&cYesNo);
		cYesNo = tolower(cYesNo);
	} while (cYesNo!='n' && cYesNo!='y');
	return (cYesNo=='y');
}

void ask_overwrite(char *filename)
{
	if (bYes == FALSE && file_exits(filename))
	{
		printf("Overwrite output file %s? (y or n): ",filename);
		if(!yesno())
		{
			printf("Program terminated.\n");
			exit(0);
		}
	}
}

void file_dump(char *filename,void *data, int len)
{
	int ret;
	FILE *fOutput = fopen(filename, "wb");
	
	if (!fOutput)
	{
		fprintf (stderr, "ERROR : Unable to open output file %s.\n",filename);
		exit(0);
	}

   	ret = fwrite(data,len,1,fOutput);
	
	fclose(fOutput);
	
	if (ret != 1)
	{
		fprintf (stderr, "ERROR : Ouput file %s appears to be invalid.\n",filename);
	   	exit(0);
	}
}

char *fileextension(char *str)
{
	return (strrchr(str,'.'));	  
}

char *newfilepath(char* filepath, char* extension)
{
	size_t len1 = strlen(filepath) - strlen(fileextension(filepath)), len2 = strlen(extension);
	char *concat = (char*) malloc(len1 + len2 + 1);
	memcpy(concat, filepath, len1);
	memcpy(concat + len1, extension, len2 + 1);
	return concat;
}

int main( int argc, char *argv[] )
{
	FILE *fInput = NULL;
	FILE *fOutput = NULL;
	char *sInput = NULL;
	char *sPC = NULL;
	char *sPattern = NULL;
	char *sColor = NULL;
	char c;
	int ret;
	int bSplit = FALSE;
	long filesize = 0;

	/*
	 * - PROGRAM TITLE AND VERSION -
	 */

	printf("%s  Version %d.%d  %s\n",TITLE,VERSION,SUBVERSION, DESCRIPTION);
	printf("by %s, %d\n\n", AUTHOR, YEAR);
	
	/*
	 * - EXTRACT COMMAND LINE OPTIONS -
	 */

	while((c=getopt(argc,argv,"hsyc:p:o:i:"))!=-1)
	{
		switch (c) {
			case 'h':
	   	   	    printf("This program convert .SCR and .MLT files into .PC files\n");
	   	   	    printf("Usage : zx2cv [options] -i input -p output -c output\n");
	   	   	    printf("    -i filename : input .SCR or .MLT file\n");
	   	   	    printf("    -o filename : output .PC file\n");
	         	printf("    -p filename : output Pattern table file\n");
	   	   	    printf("    -c filename : output Color table file\n");
	   	   	    printf("    -y          : auto-answer Yes to overwrite output file if exists\n");
	   	   	    printf("    -s          : split pattern and color tables\n");
	   	   	    printf("    -h          : show this Help\n");
				help();
			case 'i':
				sInput = optarg;
				break;
 	    	case 'o':
				sPC = optarg;
				break;
 	      	case 'p':
				sPattern = optarg;
				bSplit = TRUE;
				break;
			case 'c':
				sColor = optarg;
				bSplit = TRUE;
	   	   	   	break;
 	    	case 's':
				bSplit = TRUE;
				break;	     	 	 	 
			case 'y':
				bYes = TRUE;
				break;
	   	   	case '?':
	   	   	   	if (optopt == 'i' || optopt == 'o' || optopt == 'p' || optopt == 'c')
					fprintf(stderr,"Option -%c requires an argument.\n", optopt);
				else
	   	   	   		fprintf(stderr,"Unknown option -%c.\n", optopt);
				help();
      	    default:
				abort();
				break;

	   	}
	}
	
	if (sInput == NULL) {
		if (optind<argc) {
			sInput = argv[optind];
		}
		else
		{
			printf("Missing input file specification\n");
			help();
		}
	}

	if (sPC == NULL) {
		sPC = newfilepath(sInput,".pc");
	}

	if (bSplit)
	{
		if (sPattern == NULL) {
			sPattern = newfilepath(sInput,".pattern");
		}
	
		if (sColor == NULL) {
			sColor = newfilepath(sInput,".color");
		}
	}

	fInput = fopen(sInput, "rb");
	if (!fInput)
	{
		fprintf (stderr, "ERROR : Unable to open input file %s.\n",sInput);
		exit(0);
	}
	
	filesize = file_size(fInput);
	/* If the source is a valid file, it should fit one of the two filesize option */
	switch (filesize)
	{
		case 6912:
			/* SCR */
			ret = fread(&zx_bitmap_scr,sizeof(zx_bitmap_scr),1,fInput);
			scr2mlt();
			break;
		case 12288:
			/* MLT */
			ret = fread(&zx_bitmap_mlt,sizeof(zx_bitmap_mlt),1,fInput);	    		
			break;
		default:
			fprintf (stderr, "ERROR : Unexpected input file %s.\n",sInput);
			exit(0);	   	   
	}
	
	fclose(fInput);

	if (ret != 1)
	{
		fprintf (stderr, "ERROR : Input file %s appears to be invalid.\n",sInput);
	   	exit(0);
	}
	
	mlt2cv();

	ask_overwrite(sPC);
	file_dump(sPC,&cv_bitmap,sizeof(cv_bitmap));

	if (sPattern != NULL) {
		ask_overwrite(sPattern);
		file_dump(sPattern,&cv_bitmap.PATTERN,sizeof(cv_bitmap.PATTERN));
	}
	
	if (sColor != NULL) {
		ask_overwrite(sColor);
		file_dump(sColor,&cv_bitmap.COLOR,sizeof(cv_bitmap.COLOR));
	}

	return 0;
}

